#ifndef __BASEFUNCTIONSETSHELLFE_H
#define __BASEFUNCTIONSETSHELLFE_H

#include<aol.h>

using namespace aol;

namespace shellFE {

//! Inteface
template <typename DataTypeContainer, class QuadRuleType >
class UnitTriangleBaseFunctionSetInterface  {
public:
	typedef typename DataTypeContainer::RealType RealType;
	typedef typename DataTypeContainer::DomVecType DomVecType;

  UnitTriangleBaseFunctionSetInterface( ) {}

  //! \brief Returns the number of quadrature points
  int numQuadPoints() const {
    return QuadRuleType::numQuadPoints;
  }

  //! \brief Returns the quadrature weight at QuadPointa quadrature point
  inline RealType getWeight ( int QuadPoint ) const {
    return _quadRule.getWeight ( QuadPoint );
  }

  //! \brief Returns the coordinates of a quadrature point
  inline const DomVecType& getRefCoord ( int QuadPoint ) const {
    return _quadRule.getRefCoord ( QuadPoint );
  }

protected:
  QuadRuleType _quadRule;
};

//! \brief Base function set for unit triangle.
//! Unit triangle embedded in R^2 is given by the three positions (0,0), (1,0) and (0,1).
template <typename DataTypeContainer, typename QuadType, typename TriangleType>
class UnitTriangMeshBaseFunctionSetP1 : public UnitTriangleBaseFunctionSetInterface< DataTypeContainer, QuadType >  {
public:
  typedef typename DataTypeContainer::RealType RealType;
  typedef typename DataTypeContainer::DomVecType DomVecType;

protected:
  static RealType _b1   ( const DomVecType &c ) { return 1. - c[0] - c[1]; }
  static RealType _b2   ( const DomVecType &c ) { return c[0]; }
  static RealType _b3   ( const DomVecType &c ) { return c[1]; }

  typedef RealType ( *BASIS_FUNC_TYPE ) ( const DomVecType &RefCoord );
  BASIS_FUNC_TYPE _basis[3];
  
  const typename DataTypeContainer::TangentVecType _d1b, _d2b;

public:

  UnitTriangMeshBaseFunctionSetP1() : _d1b(-1.,1.,0.), _d2b(-1.,0.,1.) {
    _basis[0] = _b1;
    _basis[1] = _b2;
    _basis[2] = _b3;
  }
  
  enum { numBaseFuncs = 3 };

  void setTriangle ( const TriangleType &/*T*/ ) {}
  
  //! \brief Evaluates the gradient of a base function at RefCoord
  //! \param[in] BaseFuncNum The number of the base function
  //! \param[in] RefCoord The coordinate to evaluate the gradient in
  //! \param[out] Gradient The gradient at RefCoord
  void evaluateGradient ( int BaseFuncNum, const DomVecType &/*RefCoord*/, DomVecType &Gradient ) const {
    Gradient[0] = _d1b[BaseFuncNum];
    Gradient[1] = _d2b[BaseFuncNum];
  }

  //! \brief Evaluates the gradient of a base function in a quadrature point
  //! \param[in] BaseFuncNum The number of the base function
  //! \param[in] QuadPoint Number of the quadrature point to evaluate the gradient in
  //! \returns The Gradient at the quadrature point
  inline DomVecType evaluateGradient ( int BaseFuncNum, int /*QuadPoint*/ ) const {
    return DomVecType( _d1b[BaseFuncNum], _d2b[BaseFuncNum] );
  }

  //! \brief Evaluates a base function at
  //! \param[in] BaseFuncNum The number of the base function
  //! \param[in] RefCoord The coordinate to evaluate the gradient in
  //! \returns The base function value at RefCoord
  RealType evaluate ( int BaseFuncNum, const DomVecType &RefCoord ) const {
    return _basis[BaseFuncNum] ( RefCoord );
  }

  //! \brief Evaluates a base function in a quadrature point
  //! \param[in] BaseFuncNum The number of the base function
  //! \param[in] QuadPoint Number of the quadrature point to evaluate the gradient in
  //! \returns The base function value at the quadrature point
  inline const RealType evaluate ( int BaseFuncNum, int QuadPoint ) const {
    return evaluate ( BaseFuncNum, this->_quadRule.getRefCoord ( QuadPoint ) );
  }
};

} // end namespace




#endif //__BASEFUNCTIONSETSHELLFE_H
